/* PPM2SPEK.c */

/* **************************************************************************
*
*	PPM to Spektrum DX5e HF-Modul by ingo_s
*
*	V1.0	25.11.2012 - first Version with optional asynch Mode
*	V2.0	29.11.2012 - only 22ms asynchron Mode
*	V2.1	13.12.2012 - add LED PPM-Error codes, add Comperator Input select
*	V2.2	14.12.2012 - add #define for Channel-Count as Modell-Match
*
****************************************************************************/

/*
  PPM to Spektrum - only for 8/12/16MHz CPU Clock!!!
  Sends DSM2 signal using DX5e Spektrum TX module
  Some parts based on the Arduino code by daniel_arg

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program. If not, see <http://www.gnu.org/licenses/>.
*/


/*
	--- Hinweise zur Schaltung: ---
	Impuls Input geht an Input capture Pin (ICP) mit Schutzwiderstand, alternativ
	kann der Comperator Input mit 1,1V Schwelle benutzt werden (#define USE_COMPERATOR_INPUT).
	ber einen Taster wird Binding und Reichweiten Mode aktiviert.
	Taster beim Einschlaten gedrckt --> Binding Mode
	Taster beim Betrieb gedrckt     --> Reichweiten Mode
	
	Es wird nur eine LED verwendet fr die Zustnde:
	normal Mode			-->	blinken im Sekundentakt
    normnal Mode, Fehler vorgekommen --> blinken im Sekundentakt, aber flackernd
	Binding Mode		--> schnell blinken
	Reichweiten Mode	--> langsam blinken
	PPM Input Fehler	--> hektisch blinken

	Hinweis zum Timing:
	Wegen Problemen mit dem Spektrum Sende-Zylus Timing nicht nur bei mehr als 23ms, sondern auch
	bei ca. 18ms wird ab V2.0 wird nur noch das asynchrone 22ms Spektrum Timing benutzt.

	Je nach Senderfabrikat unterscheidet sich die Varianz des Servo-Impulses gegenber Spektrum
	Fr MPX ist deshalb eine Umrechnung 80% MPX auf 100% Spektrum Servoweg integriert.
	Fr eine Umrechnung bei anderen Herstellern liegen noch keine Werte vor.
	ber Tabellen ist ein Mapping und die Invertierung von Kanlen vorgesehen.
*/

/*
	--- Hinweise zur Software: ---
	Um mglichst geringen Jitter zu haben, wird hier das PPM Timing mit 125ns Auflsung gemessen,
	mit anschliessender Jitter Filterung. Ziel ist, den Ausgangs-Wert an Spektrum mit [1us] Einheiten
	mglichst Jitterfrei zu halten, da bei 100% Weg nur 682 Schritte Auflsung vorhanden sind.
	Das PPM Eingangs-Signal wird auf Fehler mittels berprfung der Anzahl Kanle und deren 
	Impulsbreite berprft. Wird die Anzahl der Kanle gendert, so muss neu eingeschaltet werden!
	Im Fehlerfall werden die letzten gltigen Werte in Verbindung mit Motor-Aus nach Ablauf von
	3 PPM-Zyklen als Hold-Time gesendet.
*/

/*
	---- Hinweis zu Daten der DX5e beim Binden ----
	Die Daten beim Binden der DX5e sind unabhngig von der momentanen Knppelposition
	0x90, 0x00 -- 0x00, 0x00;  0x05, 0xFF;  0x09, 0xFF;  0x0D, 0xFF;  0x10, 0xAA;  0x14, 0xAA 
	Eigentlich, nach Spektrum Angaben sollten sie flexibel sein, da beim Binden die Failsave Position
	angegeben wird. Das macht aber bei einigen BNF Produkten (mCPX V1) scheinbar Probleme.
*/

#include <avr/io.h>          
#include <avr/interrupt.h>
#include <stdlib.h>
#include <avr/pgmspace.h>
#include <util/delay.h>
#include <avr/wdt.h>

// ---- Anlage auswhlen -----
#define MPX_TIMING

// ---- bei kleinen PPM-Signalen (MPX Royal) den Comperator AIN1 Input benutzen -----
#define USE_COMPERATOR_INPUT              // mit 1,1V Schwelle fr kleine Signale

// ---- Option Modellnummer = Anzahl Kanle (MPX Royal) -----------
#define CHANNELCOUNT_AS_MODELLMATCH		// sinnvoll nur, wenn Sender unterschiedliche Anzahl ausgeben kann 

// ------ Projekt Settings ------
//#define TEST_HARDWARE
//#define TEST_PIN
// =================================================================================================
// ------------------------------------------ Settings ---------------------------------------------
#define WATCHDOG_ON							// 

/* TX DSM header info
 bit 7 - 1 -> bind mode enable
 bit 6 - unknown, always 0
 bit 5 - 1 -> range test enable
 bit 4 - 1 -> normal mode, 0 -> france mode
 bit 3 - 1 -> DSMX enable
 bit 2 - unknown, always 0
 bit 1 - unknown, always 0
 bit 0 - unknown, always 0
*/

// ------ DSM2 TX Modul header Codes ------
//#define CODE_NORMAL_SENDING	0x10		// DX5e Codes DSM2 only
//#define CODE_BINDING			0x90
//#define CODE_LOW_POWER		0x30

// ------ DSMX TX Modul header Codes ------
#define	CODE_NORMAL_SENDING		0x18		// DX5e Codes DSMX/DSM2
#define CODE_BINDING			0x98
#define CODE_LOW_POWER			0x38

// --------------------------------------------------------------------------------------------------
// don't change this
#define CENTER_DSM2		511					// DSM2 HF-Modul Wert fr Servo-Mitte
#ifdef USE_COMPERATOR_INPUT                 // because of inverted Comperator
#define FALLING_EDGE    (1<<ICES1)          // Edge select must be inverted
#define RISING_EDGE     (0<<ICES1)
#else
#define FALLING_EDGE	(0<<ICES1)
#define RISING_EDGE		(1<<ICES1)
#endif

// ----------------------------------- Anlagen spezifisch -------------------------------------------
// Impuls Grenzen - [us] fr die Umrechnung auf Spektrum Werte

#ifdef  MPX_TIMING
// -------------------    MPX Timing      --------------------
#define EDGE_MODE		FALLING_EDGE    
// MPX +/- 100% Servoweg [550us] ==> +/- 125% Spektrum Servoweg [426us]
#define CENTER_PULSE	1500
#define ABSMIN_PULSE   	(CENTER_PULSE - 659)		// 659 = ((511 * 58) + 22) / 45
#define ABSMAX_PULSE  	(CENTER_PULSE + 659)	
#define CONV_PPM2DSM	((tppm - ABSMIN_PULSE) * 45 / 58)
// Channel Mapping
// Achtung Motor liegt auf Kanal 4, MPX belegt bei Heli aber Kanal 5 vor, Anpassen!!!
volatile uint8_t 	ChanIndex[]   = {4,1,2,3,5,6};	// DSM2 to PPM Channel Mapping
volatile uint8_t	ChanReverse[] = {0,1,0,1,0,0};	// Channel Reverse (Quer, Seite)

#else
// ----------- allgemeines Timing fr Futaba und andere Anlagen --------
#define EDGE_MODE		FALLING_EDGE
#define CENTER_PULSE	1500						// evtl. auf 1520 setzen
#define ABSMIN_PULSE   	(CENTER_PULSE - CENTER_DSM2)
#define ABSMAX_PULSE  	(CENTER_PULSE + CENTER_DSM2)
#define CONV_PPM2DSM	(tppm - ABSMIN_PULSE)
// Channel Mapping
volatile uint8_t 	ChanIndex[]   = {3,1,2,4,5,6};    // DSM2 to PPM Channel Mapping
volatile uint8_t	ChanReverse[] = {0,0,0,0,0,0};
#endif

// ***************************************************************************************************
// ===================================================================================================
// -------------- Port defines -------------
// PortB - PPM Input
#define bIMPULS			PB0					// ICP Input an ATMegax8
// PortD
#ifndef TEST_HARDWARE
#define PIN_BIND		PIND				// Bind Taster Input Port
#define bBIND     		PD5               	
#define PORT_LED		PORTD				// LED Output Port
#define bLED      		PD6                
#else	// else Test auf Lipo-Saver
#define PIN_BIND		PIND				// Bind Taster Input Port
#define bBIND     		PD4               	
#define PORT_LED		PORTD				// LED Output Port
#define bLED      		PD3                
#endif
// Test-Pin fr Test Ausgaben
#ifdef TEST_PIN
#define bTest	 PB1			
#define TESTPORT PORTB
#define TESTDDR	 DDRB
#define SetTestPin TESTPORT |= 1<<bTest;
#define ResetTestPin TESTPORT &= ~(1<<bTest);
#else 
#define SetTestPin
#define ResetTestPin
#endif 

// ----------------- interne feste Werte ---------------------
#define TICKS_PER_uS  (F_CPU/1000000L)     // number of timer ticks per 1us 
#define DE_JITTER     ((TICKS_PER_uS/2)+1)
// Impuls Grenzen - physikalisch zulssig, auerhalb Error-Status!
#define MIN_IN_PULSE  ( 750 * TICKS_PER_uS) // minimale Impulsbreite  750us
#define MAX_IN_PULSE  (2250 * TICKS_PER_uS) // maximale Impulsbreite 2250us
#define SYNC_GAP_LEN  (3500 * TICKS_PER_uS) // 3500us als minimale Synch Pause
// Anzahl Channels
#define MAX_CHANNELS      9                 // maximum number of ppm input channel
#define DSM2_CHANNELS     6                 // Max number of DSM2 Channels transmitted
#define VALID_FRAMES      8                 // power on number of good frames for going online

// ----- OnBord Status LED -----
#define LEDon	PORT_LED |=  (1<<bLED);
#define LEDoff	PORT_LED &= ~(1<<bLED);

// ----- UART ------
#define BAUD 125000
#define MYUBRR F_CPU/16/BAUD-1

// ----- PPM Status ----- 
typedef enum {
    NULL_ST = -1, NOT_SYNCHED, ACQUIRING, READY, FAILSAFE
} State_t;
// ----- Sende Mode -----
typedef enum {
	NULL_MD = -1, SEND_DSM2, BINDING, LOW_PWR
} Mode_t;

// Watchdog defines
#ifdef WATCHDOG_ON
#define WDReset		wdt_reset();	// wdt reset
#else
#define	WDReset		//no wdt_reset
#endif

// ---------------------------------------- Functions -----------------------------------------------
void Init_USART( unsigned int ubrr);
void usart_send_char( uint8_t data );
void sendDSM2(void);

// ------------------------------------------ Data --------------------------------------------------
volatile uint16_t	rawImpVal[MAX_CHANNELS+1];		// array Impuls Rohwerte mit 8 tics/us
volatile uint16_t 	Pulses[  MAX_CHANNELS + 1];		// array mit auf DSM2 Wert umgerechneten Impulsbreiten
volatile uint16_t 	Failsafe[MAX_CHANNELS + 1];     // array dito fr FAILSAVE Status
volatile uint8_t 	ChannelNum = 0;                 // number of channels detected so far in the frame (first channel is 1)
volatile uint8_t 	ChannelCnt = 0;                 // the total number of channels detected in a complete frame
volatile State_t 	State;                   	    // this will be one of the following states: Null, Not_Synched, Acquiring, Ready, Failsafe
volatile Mode_t		Mode;							// TX-Mode null, send_dsm2, binding, low_power
volatile uint8_t 	stateCount = 0;                 // counts the number of valid frames on ACQUIRING

struct DSM2_Data {									// TX Daten array
	uint8_t Header[2];
	uint8_t	Channel[DSM2_CHANNELS*2];
} DSM2_Data;

volatile uint8_t  	flash_duration ;                // flashing LED duration calc
volatile uint8_t 	DSM2_Sent;						// Sende Anforderungs Flag
volatile uint8_t	ppmERROR;						// store PPM Failsave for ever
// ------------------------------------------ ISR --------------------------------------------------

static volatile uint8_t	cntCompaOvf;			

// ---- PPM timeout check, used for Frame-Time Measurement too -----
ISR(TIMER1_COMPA_vect) {			// on Frame-Start cntCompaOvf=0 and OCR1A=ICR1
	if (cntCompaOvf > 8){				
	  	if(State == READY) {
	    	State = FAILSAFE;       // use fail safe values if signal lost
			ppmERROR = 4;			// remember Failsave
	    	ChannelNum = 0;         // reset the channel count
	  	}   
	}
	else
		cntCompaOvf++;                                        
}                                           

// inline - Sync Function
static void processSync() {                 // sync pulse was detected so reset the channel to first and update the system state
  if(State == READY) {                      
    if( ChannelNum != ChannelCnt){           // if the number of channels is unstable, go into failsafe
      State = FAILSAFE; 
	  ppmERROR = 4;							// remember Failsave
	}                    
  }                                         
  else {                                    
    if(State == NOT_SYNCHED) {              
      State = ACQUIRING;                    // this is the first sync pulse, we need one more to fill the channel data array
      stateCount = 0;                       
    } else {                                
      if(State == ACQUIRING) {             
        if(++stateCount > VALID_FRAMES) {
          State = READY;                    // this is the second sync and all channel data is ok so flag that channel data is valid
          ChannelCnt = ChannelNum;          // save the number of channels detected
#ifdef CHANNELCOUNT_AS_MODELLMATCH
		  // Test, die Anzahl Channels bestimmt die Modellnummer
		  DSM2_Data.Header[1] = ChannelCnt;
#endif
        }                                   
      } else                                
        if( State == FAILSAFE) {            
          if(ChannelNum == ChannelCnt){      // did we get good pulses on all channels?
            State = READY;                  
#ifdef CHANNELCOUNT_AS_MODELLMATCH
			DSM2_Data.Header[1] = ChannelCnt;
#endif
		  }
        }                                   
    }                                       
  }                                         
  ChannelNum = 0;                           // reset the channel counter
}


// der Timer1 luft immer durch, fr genauere Messungen mit 125ns Auflsung
ISR(TIMER1_CAPT_vect) {  
	static uint16_t	ppm_ltime;				// last capture time
	uint16_t tdiff, tppm=0;         

	uint16_t ppm_atime = ICR1;				// actual capture time

SetTestPin
	// CompA Overflow oder Pause > SYNC_GAP_LEN -> Synchronisierung
  	if( cntCompaOvf || (ppm_atime - ppm_ltime) >= SYNC_GAP_LEN){                  // is the space between pulses big enough to be the SYNC
	    processSync();  
	}
  	else{                                      
	    if(ChannelNum < MAX_CHANNELS) {         // check if its a valid channel pulse and save it
			tdiff = ppm_atime - ppm_ltime;
	      	if((tdiff >= MIN_IN_PULSE) && (tdiff <= MAX_IN_PULSE)){  // check for valid channel data
				ChannelNum++;
				// Impuls-Jitter etwas filtern
				tppm =rawImpVal[ChannelNum];  
				if (tdiff > tppm+DE_JITTER){
					tppm = tdiff-DE_JITTER;
					rawImpVal[ChannelNum] = tppm;
				}
				else if (tdiff < tppm-DE_JITTER){
					tppm = tdiff+DE_JITTER;
					rawImpVal[ChannelNum] = tppm;
				}
			}
	      	else{
		        if(State == READY) {
		          State = FAILSAFE;                 // use fail safe values if input data invalid
				  ppmERROR = 4;						// remember Failsave
		          ChannelNum = 0;                   // reset the channel count
		        }
			}
			if(State == READY){
				tppm = tppm / TICKS_PER_uS;			// umrechnen auf [us] Einheiten
				// numerischen min/max Bereich eingrenzen
				if (tppm < ABSMIN_PULSE)
					tppm = ABSMIN_PULSE;
				else if (tppm > ABSMAX_PULSE)
					tppm = ABSMAX_PULSE;
				// aktuellen, begrenzten Wert umrechnen und speichern
				tppm = CONV_PPM2DSM;					// Umrechnung per #define 
				Pulses[ChannelNum] = tppm ;	
			}
	    }
	}
	// beim letzten Impuls haben wir Zeit alle Impulswerte fr den PPM-Error zu bernehmen
	if((State == READY) && (ChannelCnt == ChannelNum)) {	// nur bei Frames ohne Fehler
		for (uint8_t i=1; i<=ChannelCnt; i++) 				// loop only aktiv Channels 
			Failsafe[i] = Pulses[i];		
	}
	// immer letzten aktuellen ICP Zeitpunkt festhalten
	ppm_ltime = ppm_atime;
	// reset Synchronisierung berwachung
	cntCompaOvf = 0;				// reset Overflow counter
	OCR1A = ppm_atime;				// update Compare Register	
	TIFR1 = (1<<OCF1A);				// clear CompA isr flag

ResetTestPin
}

#define COMB_LOOPS  ((22000UL*TICKS_PER_uS)/65536L)
#define COMP_ADJ    ((22000UL*TICKS_PER_uS) - COMB_LOOPS*65536L)
// OCR1B macht das 22ms Spektrum Timing
ISR(TIMER1_COMPB_vect) {	
	static uint8_t	cntCompbOvf=0;

	// fr 22ms sind 2 volle + eine Teilrunde notwendig (8MHz)
	if (++cntCompbOvf == COMB_LOOPS)
		OCR1B += COMP_ADJ;			// (22000ms * 8) - 2*65536
	else if (cntCompbOvf > COMB_LOOPS){
		cntCompbOvf = 0;
		DSM2_Sent = 1;				// Senden!
	}
}


// ------------------------------------------ Init -------------------------------------------------

static void InitData(void)
{
    State   = NOT_SYNCHED; 
	Mode	= SEND_DSM2;    
	ppmERROR = 0;
	// set midpoint as default values for pulses and failsafe
	for(uint8_t ch = 0; ch <= MAX_CHANNELS; ch++) {
		Failsafe[ch] = Pulses[ch] = CENTER_DSM2;     
	}
}

static void InitPorts(void)
{
	// ---- on PortB only Impuls Input ----
	DDRB = 0;					// all Input
	PORTB = (1<<bIMPULS);		// with PullUp on ImpulsIn
	// ---- on PortD LED and SwitchButton ----
	DDRD = 0;
	PORTD = (1<<bBIND);			// set PullUp
	DDRD |= (1<<bLED);			// Output for LED

#ifdef TEST_PIN
	TESTDDR |= (1<<bTest);
#endif
}

static void InitTimer(void)
{
#ifdef USE_COMPERATOR_INPUT
    ACSR = (1<<ACBG)|(1<<ACIC);					// use AIN0 not ICP Input
#endif
	// 16bit Timer1 wird mit F_CPU Clock betrieben und luft immer rund
	TCCR1A	= 0;								// Normal port operation, OC1A/OC1B disconnected
	TCCR1B	= (1<<ICNC1)|(1<<CS10)|EDGE_MODE;		// icp noise filter,falling edge, clk/1 = 8tics/us
    TIMSK1  = (1<<ICIE1)|(1<<OCIE1A)|(1<<OCIE1B);   // enable input capture and output compareA+B
}


// -------------------------------------------------------------------------------------------------
// ++++++++++++++++++++++++++++++++++++++++++ MAIN +++++++++++++++++++++++++++++++++++++++++++++++++


int main(void)
{
	volatile uint8_t	flash_count;		// LED flashing counter

	// hier bei Bedarf noch Resetcondition prfen

	InitData();
	Init_USART(MYUBRR);
	InitPorts();							// Ports, timer usw. initialisieren

	LEDon
	_delay_ms(100);
	LEDoff

#ifdef WATCHDOG_ON							// wdt ist normalerweise freigeschaltet
	wdt_enable(WDTO_250MS);
#endif

	if ((PIN_BIND & (1<<bBIND)) == 0)		// Binding aktivieren, wenn Taster gedrckt ist
		Mode = BINDING;

	InitTimer();
	// Interrupt freischalten
	sei();	

  	uint8_t count = 100;
  	while(State != READY && count-- > 0){ // wait 5 sec or until PPM data is stable and ready
    	_delay_ms(50);
		WDReset
	}

	DSM2_Sent = 0;

	while(1)
	{
		// ------ Taster Abfrage -----
		// wenn Taster nicht gedrckt, ist immer SEND_DSM2 Mode
		if ( PIN_BIND & (1<<bBIND) )				
			Mode = SEND_DSM2;						
		// bei gedrcktem Schalter nur von SEND_DSM2 auf Reichweiten-Test umschalten
		if ( (Mode == SEND_DSM2) && !(PIN_BIND & (1<<bBIND)) )
			Mode = LOW_PWR;
		// ---------------------------

		if(State == FAILSAFE)							// bei PPM Fehler hektisch blinken
			flash_duration = 4;

      	if((flash_count%flash_duration) < flash_duration/2)  // flash LED routine
			if (ppmERROR){
				if ((flash_count%4) < 3)
					LEDon else LEDoff
			}
			else
				LEDon
		else
			LEDoff

	    if(State == READY || State == FAILSAFE) {
	        if(DSM2_Sent == 1) {                  	// if DSM2 frame is not sent yet
		        sendDSM2();                         // send frame
				flash_count++;
			}
		}

	} //while(1)

	return 0;
}

// ---------------------------------------- Functions --------------------------------------------------
// --- Spektrum channel bezogene Werte mit Mapping auf die Input Daten holen ---
uint16_t getChannelData(uint8_t channel) // this is the access function for channel data
{     
	uint16_t result = 85;           				// Failsafe -125% MotorAus default value

    if(State == READY) {
      cli();                            			// disable interrupts
      result = Pulses[ChanIndex[channel]];  		// return the last valid pulse width for this channel
      sei();                            			// enable interrupts
    }
	else{	// (State == FAILSAVE)
		if (channel > 0 ){			// immer Data-Hold auer Motor
			cli();                            		// disable interrupts
		    result = Failsafe[ChanIndex[channel]];  // return the channels failsafe value if set and State is Failsafe
			sei();                            		// enable interrupts
		}
		else{
			cli();                            		// disable interrupts
			if (ppmERROR > 1){		// Motor Hold counter
				ppmERROR--;
			    result = Failsafe[ChanIndex[channel]];  // return the channels failsafe value if set and State is Failsafe
			}
			sei();                            		// enable interrupts
		}
	}

	// Servo-Reserve, weil Spektrum andere default Richtungen hat
	if (ChanReverse[channel])
		result = 1023 - result;

	return result;
}


// --- aktuelle Channel Werte und Mode an HF-Modul bertragen ---
void sendDSM2(void)
{
	WDReset
	if (Mode == SEND_DSM2){
      	DSM2_Data.Header[0] = CODE_NORMAL_SENDING;	// set normal ops header
		flash_duration = 200;						// ganz langsam blinken
	}
	else if (Mode == BINDING){
  		DSM2_Data.Header[0] = CODE_BINDING;			// set binding ops header
		flash_duration = 20;  
	}
	else{
  		DSM2_Data.Header[0] = CODE_LOW_POWER;       // set low power ops header
		flash_duration = 40;
	}
	// das knnte die Modellspeicher Nummer sein fr Modellmatch, einfach ausprobieren
#ifndef CHANNELCOUNT_AS_MODELLMATCH
  	DSM2_Data.Header[1] = 0x00;              		// second header byte  
#endif
	// alle 6 Channel Werte holen
	for (uint8_t i=0; i<DSM2_CHANNELS; i++) {	// get receiver data
		uint16_t temp = getChannelData(i);
		DSM2_Data.Channel[i*2]   = (uint8_t)(i<<2) | (temp>>8);
		DSM2_Data.Channel[i*2+1] = temp;
	}

	// die 2 HeaderBytes und die Channel-Werte senden
	for (uint8_t i=0; i<(sizeof(DSM2_Data)); i++)
	{
		usart_send_char(DSM2_Data.Header[i]);
	}

	DSM2_Sent = 0; 
}

// --- USART Funktionen aus dem Datenblatt ---
void Init_USART( unsigned int ubrr)
{
	/* Set baud rate */
	UBRR0H = (unsigned char)(ubrr>>8);
	UBRR0L = (unsigned char)ubrr;
	/* Enable receiver and transmitter */
	UCSR0B = (1<<RXEN0)|(1<<TXEN0);
	/* Set frame format: 8data, 2stop bit */
	UCSR0C = (1<<USBS0)|(3<<UCSZ00);
}

void usart_send_char( uint8_t data )
{
	/* Wait for empty transmit buffer */
	while ( !( UCSR0A & (1<<UDRE0)) )
	;
	/* Put data into buffer, sends the data */
	UDR0 = data;
}


// -------------------------------------------------------------------------------------------------
